/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.ConfigUtil;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.persistence.impl.Util;
import com.topcoder.document.index.wordsource.WordSourceId;

import com.topcoder.util.config.Property;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import java.text.CollationKey;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;


/**
 * <p>
 * This is a concrete implementation of the IndexPersistence contract that uses xml files as storage medium. This
 * implementation uses a dual strategy - SAX for reading data in and DOM for writing data out. It manages three kinds
 * of files:
 *
 * <ul>
 * <li>
 * document index file (doc_index_docId_sequenceId.xml) - stores the information for one document index. May be split
 * to several sequence files.
 * </li>
 * <li>
 * collection index file (doc_collection_collectionId.xml) - stores the information for one collection index.
 * </li>
 * <li>
 * document usage file (document_usage.xml) - stores the relationship about which collection indexes each document
 * index belongs to.
 * </li>
 * </ul>
 *
 * The setup requirement of the xml persistence is to have the document usage file (document_usage.xml) under the
 * persistence directory. On the first use, the file should contain the documents tag only.
 * </p>
 *
 * <p>
 * It supports the loading of xml-related properties from configuration file. Here is a sample:
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;CMConfig&gt;
 *   &lt;Config name="com.topcoder.document.index.persistence.impl.xml"&gt;
 *     &lt;!-- the file size limit for saving a chunk of indexed data, in Mbytes (required) --&gt;
 *     &lt;Property name="FileSizeLimit"&gt;
 *       &lt;Value&gt;2000&lt;/Value&gt;
 *     &lt;/Property&gt;
 *     &lt;!-- the safety net size that is applied before the FileSizeLimit is reached, in Kbytes (required) --&gt;
 *     &lt;Property name="FileSizeThreshold"&gt;
 *       &lt;Value&gt;10&lt;/Value&gt;
 *     &lt;/Property&gt;
 *     &lt;!-- the file path to a directory where all the xml index files are stored (required) --&gt;
 *     &lt;Property name="XmlPersistencePath"&gt;
 *       &lt;Value&gt;c:\xmldata&lt;/Value&gt;
 *     &lt;/Property&gt;
 *   &lt;/Config&gt;
 * &lt;/CMConfig&gt;
 * </pre>
 * </p>
 *
 * <p>
 * This class is not thread-safe. The methods should not be called concurrently on a single instance. Unlike database,
 * the xml files do not support atomic transactions (commit and rollback). Multiple instances of this class should not
 * access the persistence at the same time.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistence implements IndexPersistence {
    /**
     * <p>
     * The file name used to represent a document usage as defined by the persistence contract. There is only one such
     * file per xml persistence namespace.
     * </p>
     */
    private static final String DOC_USAGE_FILE_NAME = "document_usage.xml";

    /**
     * <p>
     * The file name prefix for a collection index file. Each collection index will have the index file of this format:
     * doc_collection_collectionId.xml. The collectionId is the id of the collection index.
     * </p>
     */
    private static final String DOC_COLLECTION_FILE_NAME_PREFIX = "doc_collection_";

    /**
     * <p>
     * The file name prefix for a document index file. Each document index will have the index file of this format:
     * doc_index_docId_sequenceId.xml. The docId is the id of the document index. The sequenceId is used to link files
     * with data that is larger than fileSizeLimit. The sequenceId starts with 1 and grows sequentially.
     * </p>
     */
    private static final String DOC_INDEX_FILE_NAME_PREFIX = "doc_index_";

    /**
     * <p>
     * The file extension for all files managed by this persistence.
     * </p>
     */
    private static final String FILE_EXTENSION = ".xml";

    /**
     * <p>
     * Constant for the property specifying the file size limit of the document index file.
     * </p>
     */
    private static final String FILE_SIZE_LIMIT_PROPERTY = "FileSizeLimit";

    /**
     * <p>
     * Constant for the property specifying the file size threshold of the document index file.
     * </p>
     */
    private static final String FILE_SIZE_THRESHOLD_PROPERTY = "FileSizeThreshold";

    /**
     * <p>
     * Constant for the property specifying the persistence directory path. All files managed by this persistence
     * reside in this directory.
     * </p>
     */
    private static final String XML_PERSISTENCE_PATH_PROPERTY = "XmlPersistencePath";

    /**
     * <p>
     * The file size limit for saving a chunk of document indexed data. It is measured in Mbytes. In some cases the
     * data cannot fit into a single file due to size limitations of the specific OS. This size limit is used to
     * decide when to start chaining multiple files to store the data. It is initialized from the constructor through
     * configuration and cannot be changed. Must be a positive integer.
     * </p>
     */
    private final int fileSizeLimit;

    /**
     * <p>
     * The safety net size in Kbytes that is applied before the fileSizeLimit is reached. It is initialized from the
     * constructor through configuration and cannot be changed. Must be a non-negative integer.
     * </p>
     */
    private final int fileSizeThreshold;

    /**
     * <p>
     * The file path to a directory where all the xml index files are stored. This must be a valid absolute path. It is
     * initialized from the constructor through configuration and cannot be changed. Cannot be null or an empty
     * string.
     * </p>
     */
    private final File persistenceDirFile;

    /**
     * <p>
     * A constructor that loads the xml-related properties from the specified namespace.
     * </p>
     *
     * @param namespace The namespace for xml-related properties.
     *
     * @throws IllegalArgumentException if namespace is null or empty string.
     * @throws PersistenceConfigurationException if there are configuration issues encountered.
     */
    public XmlIndexPersistence(String namespace) throws PersistenceConfigurationException {
        if (namespace == null) {
            throw new IllegalArgumentException("namespace is null");
        }
        if (namespace.trim().length() == 0) {
            throw new IllegalArgumentException("namespace is empty string");
        }

        Property root = ConfigUtil.getRootProperty(namespace);

        // read the required properties
        String persistenceDirPath = ConfigUtil.getPropertyValue(root, XML_PERSISTENCE_PATH_PROPERTY);

        fileSizeLimit = ConfigUtil.getIntPropertyValue(root, FILE_SIZE_LIMIT_PROPERTY);
        fileSizeThreshold = ConfigUtil.getIntPropertyValue(root, FILE_SIZE_THRESHOLD_PROPERTY);

        try {
            // check the parameters
            checkParameters(fileSizeLimit, fileSizeThreshold, persistenceDirPath);
        } catch (IllegalArgumentException e) {
            // re-throw it as PersistenceConfigurationException
            throw new PersistenceConfigurationException(e.getMessage());
        }

        // setup the directory if everything is fine
        persistenceDirFile = new File(persistenceDirPath);
    }

    /**
     * <p>
     * A constructor that accepts the file size limit (in Mbytes), file size threshold (in Kbytes) and persistence
     * directory path. The directory path must be an absolute path, pointing to an existing and writable directory
     * containing the document usage file.
     * </p>
     *
     * @param fileSizeLimit the file size limit for document index file, in Mbytes.
     * @param fileSizeThreshold the safety net size that is applied before the fileSizeLimit is reached, in Kbytes.
     * @param persistenceDirPath the file path to a directory where all the xml index files are stored.
     *
     * @throws IllegalArgumentException if fileSizeLimit is non-positive, fileSizeThreshold is negative, or
     *         persistenceDirPath is null or empty string, or not a valid directory path.
     */
    public XmlIndexPersistence(int fileSizeLimit, int fileSizeThreshold, String persistenceDirPath) {
        // check the parameters
        checkParameters(fileSizeLimit, fileSizeThreshold, persistenceDirPath);

        // setup the directory if everything is fine
        this.fileSizeLimit = fileSizeLimit;
        this.fileSizeThreshold = fileSizeThreshold;
        persistenceDirFile = new File(persistenceDirPath);
    }

    /**
     * <p>
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails.
     * </p>
     *
     * <p>
     * This method creates a new document index file (splitting it into sequence files if necessary), and adds a record
     * to the document usage file.
     * </p>
     *
     * @param documentIndex the document index to be persisted.
     *
     * @throws IllegalArgumentException if documentIndex is null.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *         persistence.
     */
    public void addDocumentIndex(DocumentIndex documentIndex)
        throws IndexPersistenceException {
        Util.checkDocumentIndex(documentIndex);

        // get the document id from the word source id
        String documentId = Util.getDocumentId(documentIndex.getWordSourceId());

        // check if the document index does not exist
        if (existDocumentIndex(documentId)) {
            throw new IndexPersistenceException("Document index already exists");
        }

        // create the document index file
        createDocumentIndexFile(documentId, documentIndex);

        // use DOM parsing to perform update of document usage file
        File file = getDocumentUsageFile();
        Document document = XmlUtil.parse(getDocumentUsageFile());

        Element documents = document.getDocumentElement();

        // create a document usage element with document id and usage count attributes
        Element documentUsage = document.createElement(XmlConstants.DOCUMENT_USAGE_TAG_NAME);

        documentUsage.setAttribute(XmlConstants.DOCUMENT_ID_ATTR_NAME, documentId);
        documentUsage.setAttribute(XmlConstants.USAGE_COUNT_ATTR_NAME, "0");

        // put it under documents
        documents.appendChild(documentUsage);

        // create a collection inclusion element
        Element collectionInclusion = document.createElement(XmlConstants.COLLECTION_INCLUSION_TAG_NAME);

        // put it under document usage
        documentUsage.appendChild(collectionInclusion);

        // write the DOM back to the file
        XmlUtil.transform(file, document);
    }

    /**
     * <p>
     * Gets the document index from the persistence. The ID of the document is given. When the document is not
     * persisted in the persistence, null should be returned.
     * </p>
     *
     * <p>
     * This method reads all the sequence files belonging to the document index.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     *
     * @return the document index of the document with the ID, or null if not found.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if retrieving the document index fails.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);

        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        // check if the document index exists
        if (!existDocumentIndex(documentId)) {
            return null;
        }

        // read all the sequence files
        return getDocumentIndex(documentId);
    }

    /**
     * <p>
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted,
     * it fails.
     * </p>
     *
     * <p>
     * This method deletes the document index file, and a record from the document usage file.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *         or there is at least one collection containing the document index.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);

        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        // check if the documnt index exists and is unused
        DocumentUsageParser parser = new DocumentUsageParser();

        XmlUtil.parse(getDocumentUsageFile(), parser);

        if (parser.getUsageCount(documentId) != 0) {
            throw new IndexPersistenceException("Unused document index does not exist");
        }

        // delete the document index sequence files
        for (int sequence = 1;; sequence++) {
            File file = getDocumentIndexFile(documentId, sequence);

            if (!file.exists()) {
                break; // stop when the sequence breaks
            }
            file.delete();
        }

        // use DOM parsing to perform update of document usage file
        File file = getDocumentUsageFile();
        Document document = XmlUtil.parse(getDocumentUsageFile());

        // locate the document usage of matched document id
        Node[] nodes = XmlUtil.toNodeArray(document.getElementsByTagName(XmlConstants.DOCUMENT_USAGE_TAG_NAME));

        for (int i = 0; i < nodes.length; i++) {
            Element documentUsage = (Element) nodes[i];

            if (documentId.equals(documentUsage.getAttribute(XmlConstants.DOCUMENT_ID_ATTR_NAME))) {
                // remove it from documents
                documentUsage.getParentNode().removeChild(documentUsage);
                break;
            }
        }

        // write the DOM back to the file
        XmlUtil.transform(file, document);
    }

    /**
     * <p>
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails.
     * When the document indexes of the collection index are not persisted, it also fails.
     * </p>
     *
     * <p>
     * This method creates a new collection index file, and add records to the document usage file. The use count is
     * not updated here, since it is handled by the CollectionIndex class.
     * </p>
     *
     * @param collectionIndex the collection index to be persisted.
     *
     * @throws IllegalArgumentException if collectionIndex is null.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in
     *         the persistence; or the document indexes of the collection index are not persisted.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        Util.checkCollectionIndex(collectionIndex);

        // get the collection id from the collection index
        String collectionId = collectionIndex.getId();

        // check if the collection index does not exist
        if (existCollectionIndex(collectionId)) {
            throw new IndexPersistenceException("Collection index already exists");
        }

        Set documentIds = Util.getDocumentIds(collectionIndex.getAllDocumentIds());

        // check if all document indexes in the collection index exist
        if (!existDocumentIndexes(documentIds)) {
            throw new IndexPersistenceException("Some document index does not exist");
        }

        // create the collection index file
        createCollectionIndexFile(collectionId, documentIds);

        // use DOM parsing to perform update of document usage file
        File file = getDocumentUsageFile();
        Document document = XmlUtil.parse(getDocumentUsageFile());

        // locate the document usage of any matched document id
        Node[] nodes = XmlUtil.toNodeArray(document.getElementsByTagName(XmlConstants.DOCUMENT_USAGE_TAG_NAME));

        for (int i = 0; i < nodes.length; i++) {
            Element documentUsage = (Element) nodes[i];

            if (documentIds.contains(documentUsage.getAttribute(XmlConstants.DOCUMENT_ID_ATTR_NAME))) {
                // create a collection element with collection id attribute
                Element collection = document.createElement(XmlConstants.COLLECTION_TAG_NAME);

                collection.setAttribute(XmlConstants.COLLECTION_ID_ATTR_NAME, collectionId);

                // put it under collection inclusion
                documentUsage.getFirstChild().appendChild(collection);
            }
        }

        // write the DOM back to the file
        XmlUtil.transform(file, document);
    }

    /**
     * <p>
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, null should be returned.
     * </p>
     *
     * <p>
     * This method reads the collection index file, and the document index files belonging to the collection index.
     * Note that the collection index is created with null owner. It should be set by the IndexManager using the
     * setOwner() method.
     * </p>
     *
     * @param collectionId the ID of the collection index to be retrieved.
     *
     * @return the collection index with the given ID.
     *
     * @throws IllegalArgumentException if collectionId is null or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        Util.checkCollectionId(collectionId);

        // check if the collection index exists
        if (!existCollectionIndex(collectionId)) {
            return null;
        }

        // use SAX parser to ensure efficiency
        CollectionIndexParser parser = new CollectionIndexParser();

        XmlUtil.parse(getCollectionIndexFile(collectionId), parser);

        // extract the document ids from collection index file
        Set documentIds = parser.getAllDocumentIds();

        Map words = new HashMap();
        Set wordSourceIds = new HashSet();

        for (Iterator i = documentIds.iterator(); i.hasNext();) {
            // build the document index for each document id
            DocumentIndex documentIndex = getDocumentIndex((String) i.next());
            WordSourceId wordSourceId = documentIndex.getWordSourceId();

            // add the word source id to the set
            wordSourceIds.add(wordSourceId);

            for (Iterator j = documentIndex.getWords().keySet().iterator(); j.hasNext();) {
                CollationKey key = (CollationKey) j.next();
                Set value = (Set) words.get(key);

                // if the word has not been added, add a new set to the map
                if (value == null) {
                    value = new HashSet();
                    words.put(key, value);
                }

                // add the word source id to the set
                value.add(wordSourceId);
            }
        }

        // create the collection index with null owner
        // it should be set by the IndexManager using the setOwner() method
        return new CollectionIndex(null, words, wordSourceIds, collectionId);
    }

    /**
     * <p>
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method deletes the collection index file, and records from the document usage file. The use count is not
     * updated here, since it is handled by the CollectionIndex class.
     * </p>
     *
     * @param collectionId the ID of the collection index to be deleted.
     *
     * @throws IllegalArgumentException if collectionId is null or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void removeCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        Util.checkCollectionId(collectionId);

        // check if the collection index exists
        if (!existCollectionIndex(collectionId)) {
            throw new IndexPersistenceException("Collection index does not exist");
        }

        // delete the collection index file
        getCollectionIndexFile(collectionId).delete();

        // use DOM parsing to perform update of document usage file
        File file = getDocumentUsageFile();
        Document document = XmlUtil.parse(getDocumentUsageFile());

        // locate the collection of any matched collection id
        Node[] nodes = XmlUtil.toNodeArray(document.getElementsByTagName(XmlConstants.COLLECTION_TAG_NAME));

        for (int i = 0; i < nodes.length; i++) {
            Element collection = (Element) nodes[i];

            if (collectionId.equals(collection.getAttribute(XmlConstants.COLLECTION_ID_ATTR_NAME))) {
                // remove it from collection inclusion
                collection.getParentNode().removeChild(collection);
            }
        }

        // write the DOM back to the file
        XmlUtil.transform(file, document);
    }

    /**
     * <p>
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the
     * persistence, it fails.
     * </p>
     *
     * <p>
     * This method updates the collection index file, and records in the document usage file. The use count is not
     * updated here, since it is handled by the CollectionIndex class.
     * </p>
     *
     * @param collectionIndex the collection index to be updated.
     *
     * @throws IllegalArgumentException if collectionIndex is null.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *         found in the persistence; or the document indexes of the collection index are not persisted.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        Util.checkCollectionIndex(collectionIndex);

        // get the collection id from the collection index
        String collectionId = collectionIndex.getId();

        // check if the collection index exists
        if (!existCollectionIndex(collectionId)) {
            throw new IndexPersistenceException("Collection index does not exist");
        }

        // extract the document ids from the collection index
        Set newDocumentIds = Util.getDocumentIds(collectionIndex.getAllDocumentIds());
        Set oldDocumentIds = new HashSet();

        // check if all document indexes in the collection index exist
        if (!existDocumentIndexes(newDocumentIds)) {
            throw new IndexPersistenceException("Some document index does not exist");
        }

        // extract the old document ids from collection index file
        CollectionIndexParser parser = new CollectionIndexParser();

        XmlUtil.parse(getCollectionIndexFile(collectionId), parser);

        Set documentIds = parser.getAllDocumentIds();

        // update the collection index file
        createCollectionIndexFile(collectionId, newDocumentIds);

        // determine the document ids to be added and removed
        for (Iterator i = documentIds.iterator(); i.hasNext();) {
            String documentId = (String) i.next();

            if (!newDocumentIds.remove(documentId)) {
                oldDocumentIds.add(documentId);
            }
        }

        // use DOM parsing to perform update of document usage file
        File file = getDocumentUsageFile();
        Document document = XmlUtil.parse(file);

        // locate the document usage of any matched document id
        Node[] nodes = XmlUtil.toNodeArray(document.getElementsByTagName(XmlConstants.DOCUMENT_USAGE_TAG_NAME));

        for (int i = 0; i < nodes.length; i++) {
            Element documentUsage = (Element) nodes[i];
            String documentId = documentUsage.getAttribute(XmlConstants.DOCUMENT_ID_ATTR_NAME);

            // for collection id to be added
            if (newDocumentIds.contains(documentId)) {
                // create a collection element with collection id attribute
                Element collection = document.createElement(XmlConstants.COLLECTION_TAG_NAME);

                collection.setAttribute(XmlConstants.COLLECTION_ID_ATTR_NAME, collectionId);

                // put it under collection inclusion
                documentUsage.getFirstChild().appendChild(collection);
            }

            // for collection id to be removed
            if (oldDocumentIds.contains(documentId)) {
                // locate the collection of matched collection id
                Node[] collections = XmlUtil.toNodeArray(documentUsage.getElementsByTagName(
                            XmlConstants.COLLECTION_TAG_NAME));

                for (int j = 0; j < collections.length; j++) {
                    Element collection = (Element) collections[j];

                    if (collectionId.equals(collection.getAttribute(XmlConstants.COLLECTION_ID_ATTR_NAME))) {
                        // remove it from collection inclusion
                        collection.getParentNode().removeChild(collection);
                        break;
                    }
                }
            }
        }

        // write the DOM back to the file
        XmlUtil.transform(file, document);
    }

    /**
     * <p>
     * Increases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the document usage file.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);
        updateDocumentUseCount(wordSourceId, 1);
    }

    /**
     * <p>
     * Decreases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the document usage file.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);
        updateDocumentUseCount(wordSourceId, -1);
    }

    /**
     * <p>
     * Gets a set of all IDs of documents whose indexes are persisted in the persistence.
     * </p>
     *
     * <p>
     * This method reads all the records in the document usage file, and all the document index files.
     * </p>
     *
     * @return a set of all IDs of persisted document indexes.
     *
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        // use SAX parsing to ensure efficiency
        DocumentUsageParser parser = new DocumentUsageParser();

        XmlUtil.parse(getDocumentUsageFile(), parser);

        // extract all document ids from document usage file
        Set documentIds = parser.getAllDocumentIds();
        Set wordSourceIds = new HashSet();

        for (Iterator i = documentIds.iterator(); i.hasNext();) {
            String documentId = (String) i.next();
            DocumentIndexParser parser2 = new DocumentIndexParser();

            // we need to parse the first sequence file only
            XmlUtil.parse(getDocumentIndexFile(documentId, 1), parser2);

            // extract the word source id, and add it to the set
            wordSourceIds.add(parser2.getWordSourceId());
        }
        return wordSourceIds;
    }

    /**
     * <p>
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails.
     * </p>
     *
     * <p>
     * This method retrieves a record from the document usage file.
     * </p>
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     *
     * @return the reference count of the document index.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *         not persisted.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);

        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        // check if the document index exists
        if (!existDocumentIndex(documentId)) {
            throw new IndexPersistenceException("Document index does not exist");
        }

        // use SAX parsing to ensure efficiency
        DocumentUsageParser parser = new DocumentUsageParser();

        // get the usage count from the document usage file
        XmlUtil.parse(getDocumentUsageFile(), parser);

        return parser.getUsageCount(documentId);
    }

    /**
     * <p>
     * Checks the parameters for the xml file persistence. It makes sure that the persistenceDirPath points to a valid
     * directory path.
     * </p>
     *
     * @param fileSizeLimit the file size limit for document index file, in Mbytes.
     * @param fileSizeThreshold the safety net size that is applied before the fileSizeLimit is reached, in Kbytes.
     * @param persistenceDirPath the file path to a directory where all the xml index files are stored.
     *
     * @throws IllegalArgumentException if any parameter does not pass the check.
     */
    private void checkParameters(int fileSizeLimit, int fileSizeThreshold, String persistenceDirPath) {
        // check the fileSizeLimit parameter
        if (fileSizeLimit <= 0) {
            throw new IllegalArgumentException("fileSizeLimit must be positive");
        }

        // check the fileSizeThreshold parameter
        if (fileSizeThreshold < 0) {
            throw new IllegalArgumentException("fileSizeThreshold must be non-negative");
        }

        // check the resulting file size limit
        if (calculateFileSizeLimit(fileSizeLimit, fileSizeThreshold) <= 0) {
            throw new IllegalArgumentException("fileSizeLimit is too small");
        }

        // check the persistenceDirPath parameter
        if (persistenceDirPath == null) {
            throw new IllegalArgumentException("persistenceDirPath is null");
        }
        if (persistenceDirPath.trim().length() == 0) {
            throw new IllegalArgumentException("persistenceDirPath is empty string");
        }

        // make sure the persistenceDirPath points to a valid directory path
        File file = new File(persistenceDirPath);

        if (!file.isAbsolute()) {
            throw new IllegalArgumentException("persistenceDirPath is not an absolute path");
        }
        if (!file.isDirectory()) {
            throw new IllegalArgumentException("persistenceDirPath is not a directory");
        }
        if (!file.exists()) {
            throw new IllegalArgumentException("persistenceDirPath does not exist");
        }
        if (!file.canWrite()) {
            throw new IllegalArgumentException("persistenceDirPath is not writable");
        }

        // make sure the document usage file exists in the directory
        if (!new File(file, DOC_USAGE_FILE_NAME).exists()) {
            throw new IllegalArgumentException("Document usage file does not exist in persistenceDirPath");
        }
    }

    /**
     * <p>
     * Calculates the file size limit with the specified parameters. This is the maximum file size of a document file.
     * </p>
     *
     * @param fileSizeLimit the file size limit for document index file, in Mbytes.
     * @param fileSizeThreshold the safety net size that is applied before the fileSizeLimit is reached, in Kbytes.
     *
     * @return the file size limit.
     */
    private long calculateFileSizeLimit(int fileSizeLimit, int fileSizeThreshold) {
        return (fileSizeLimit * 1024000) - (fileSizeThreshold * 1000);
    }

    /**
     * <p>
     * Returns the collection index file object for the given collection id. It must be under the persistence
     * directory.
     * </p>
     *
     * @param collectionId the id of the collection index.
     *
     * @return the collection index file object.
     */
    private File getCollectionIndexFile(String collectionId) {
        return new File(persistenceDirFile, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + FILE_EXTENSION);
    }

    /**
     * <p>
     * Returns the document index file object for the given document id and sequence number. It must be under the
     * persistence directory.
     * </p>
     *
     * @param documentId the unique id of the document index.
     * @param sequence the sequence number of the file.
     *
     * @return the document index file object.
     */
    private File getDocumentIndexFile(String documentId, int sequence) {
        return new File(persistenceDirFile, DOC_INDEX_FILE_NAME_PREFIX + documentId + "_" + sequence + FILE_EXTENSION);
    }

    /**
     * <p>
     * Returns the document usage file object. It must be under the persistence directory.
     * </p>
     *
     * @return the document usage file object.
     */
    private File getDocumentUsageFile() {
        return new File(persistenceDirFile, DOC_USAGE_FILE_NAME);
    }

    /**
     * <p>
     * Gets the usage count attribute node for the given document id. The document index must exist in the persistence.
     * </p>
     *
     * @param document the document object of the document usage file.
     * @param documentId the unique id of the document index.
     *
     * @return the usage count attribute node.
     *
     * @throws IndexPersistenceException if the usage count attribute is missing or its value is not a number.
     */
    private Attr getUsageCountAttr(Document document, String documentId)
        throws IndexPersistenceException {
        Node[] nodes = XmlUtil.toNodeArray(document.getElementsByTagName(XmlConstants.DOCUMENT_USAGE_TAG_NAME));
        Attr usageCount = null;

        // locate the usage count attribute from the document id
        for (int i = 0; i < nodes.length; i++) {
            Element documentUsage = (Element) nodes[i];

            if (documentId.equals(documentUsage.getAttribute(XmlConstants.DOCUMENT_ID_ATTR_NAME))) {
                usageCount = documentUsage.getAttributeNode(XmlConstants.USAGE_COUNT_ATTR_NAME);
                break;
            }
        }

        // check if any of the fields is missing
        if (usageCount == null) {
            throw new IndexPersistenceException(XmlConstants.USAGE_COUNT_ATTR_NAME + " is missing");
        }

        // check if the usage count can be parsed as a number
        try {
            Integer.parseInt(usageCount.getValue());
        } catch (NumberFormatException e) {
            throw new IndexPersistenceException(XmlConstants.USAGE_COUNT_ATTR_NAME + " is not a number");
        }
        return usageCount;
    }

    /**
     * <p>
     * Checks whether the document index file exists. This is done by looking for the document index file in the
     * persistence directory.
     * </p>
     *
     * @param documentId the unique id of the document index.
     *
     * @return whether the document index file exists.
     */
    private boolean existDocumentIndex(String documentId) {
        // the file must always start with sequence number 1
        return getDocumentIndexFile(documentId, 1).exists();
    }

    /**
     * <p>
     * Checks whether all the document index files exists. This is done by looking for the document index files in the
     * persistence directory.
     * </p>
     *
     * @param documentIds the unique ids of the document indexes.
     *
     * @return whether all the document index files exists.
     */
    private boolean existDocumentIndexes(Set documentIds) {
        for (Iterator i = documentIds.iterator(); i.hasNext();) {
            if (!existDocumentIndex((String) i.next())) {
                return false;
            }
        }
        return true;
    }

    /**
     * <p>
     * Checks whether the collection index file exists. This is done by looking for the collection index file in the
     * persistence directory.
     * </p>
     *
     * @param collectionId the unique id of the collection index.
     *
     * @return whether the collection index file exists.
     */
    private boolean existCollectionIndex(String collectionId) {
        return getCollectionIndexFile(collectionId).exists();
    }

    /**
     * <p>
     * Gets the document index with the given document id by extracting information in the document index file. The
     * information in all sequence files is concatenated. The document index must exist in the persistence.
     * </p>
     *
     * @param documentId the unique id of the document index.
     *
     * @return the document index with the given document id.
     *
     * @throws IndexPersistenceException if retrieving the document index fails.
     */
    private DocumentIndex getDocumentIndex(String documentId)
        throws IndexPersistenceException {
        WordSourceId wordSourceId = null;
        Map words = new HashMap();

        // concatenate the information in all sequence files
        for (int sequence = 1;; sequence++) {
            File file = getDocumentIndexFile(documentId, sequence);

            if (!file.exists()) {
                break; // stop when the sequence breaks
            }

            // use SAX parsing to ensure efficiency
            DocumentIndexParser parser = new DocumentIndexParser();

            XmlUtil.parse(file, parser);

            // extract the ID of the documet index
            if (sequence == 1) {
                wordSourceId = parser.getWordSourceId();
            }

            // add the words contained in this sequence file
            words.putAll(parser.getWords());
        }

        return new DocumentIndex(wordSourceId, words);
    }

    /**
     * <p>
     * Updates the collection reference count of the document index. The document ID is given. The update argument
     * specifies the difference with the current use count. When the document index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the document usage file.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is updated.
     * @param update the difference with the current use count.
     *
     * @throws IndexPersistenceException if updating the collection reference count fails; or the document index is not
     *         persisted.
     */
    private void updateDocumentUseCount(WordSourceId wordSourceId, int update)
        throws IndexPersistenceException {
        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        // check if the document index exists
        if (!existDocumentIndex(documentId)) {
            throw new IndexPersistenceException("Document index does not exist");
        }

        // use DOM parsing to perform update of document usage file
        File file = getDocumentUsageFile();
        Document document = XmlUtil.parse(file);

        // get the usage count from the document usage file
        Attr attr = getUsageCountAttr(document, documentId);
        int usageCount = Integer.parseInt(attr.getValue());

        // update the use count
        attr.setValue(Integer.toString(usageCount + update));

        // write the DOM back to the file
        XmlUtil.transform(file, document);
    }

    /**
     * <p>
     * Creates a collection index file. The ids of the document indexes belonging to the collection index are listed.
     * </p>
     *
     * @param collectionId the id of the collection index.
     * @param documentIds the ids of the document indexes belonging to the collection index.
     *
     * @throws IndexPersistenceException if any I/O error occurs.
     */
    private void createCollectionIndexFile(String collectionId, Set documentIds)
        throws IndexPersistenceException {
        XmlWriter xw = new XmlWriter();

        // start the document collection section
        xw.writeStartElement(XmlConstants.DOCUMENT_COLLECTION_TAG_NAME, XmlConstants.COLLECTION_ID_ATTR_NAME,
            collectionId);
        xw.writeStartElement(XmlConstants.DOCUMENTS_TAG_NAME);

        // write the document ids
        for (Iterator i = documentIds.iterator(); i.hasNext();) {
            xw.writeStartElement(XmlConstants.DOCUMENT_ID_TAG_NAME);
            xw.writeText((String) i.next());
            xw.writeEndElement(XmlConstants.DOCUMENT_ID_TAG_NAME);
        }

        // end the document collection section
        xw.writeEndElement(XmlConstants.DOCUMENTS_TAG_NAME);
        xw.writeEndElement(XmlConstants.DOCUMENT_COLLECTION_TAG_NAME);

        // write the xml directly to ensure efficiency
        Writer writer = null;

        try {
            writer = new BufferedWriter(new FileWriter(getCollectionIndexFile(collectionId)));
            writer.write(xw.getString());
        } catch (IOException e) {
            throw new IndexPersistenceException("Fails to write to collection index file", e);
        } finally {
            Util.close(writer);
        }
    }

    /**
     * <p>
     * Creates a document index file. The information of the word source id as well as the words and their positions
     * are listed. It will split into several sequence files if the file size limit is reached.
     * </p>
     *
     * @param documentId the unique id of the document index.
     * @param documentIndex the document index to be persisted.
     *
     * @throws IndexPersistenceException if any I/O error occurs.
     */
    private void createDocumentIndexFile(String documentId, DocumentIndex documentIndex)
        throws IndexPersistenceException {
        // generate the header and footer of the document index file first,
        // because they are included in every file
        String header = generateDocumentIndexHeader(documentId, documentIndex);
        String footer = generateDocumentIndexFooter();
        Map words = documentIndex.getWords();

        // sequence number starts with 1
        int sequence = 1;
        long limit = calculateFileSizeLimit(fileSizeLimit, fileSizeThreshold);

        // write the xml directly to ensure efficiency
        Writer writer = null;
        XmlWriter xw = new XmlWriter();

        try {
            for (Iterator i = words.entrySet().iterator(); i.hasNext(); sequence++) {
                // create a new document index file for this sequence
                writer = new BufferedWriter(new FileWriter(getDocumentIndexFile(documentId, sequence)));

                // write the header
                writer.write(header);

                long length = header.length();

                // as long as the file size limit is not reached,
                // write the word index for next word
                while (i.hasNext() && ((length + xw.getLength()) < limit)) {
                    writer.write(xw.getString());
                    length += xw.getLength();
                    xw.reset();

                    Map.Entry entry = (Map.Entry) i.next();
                    String word = ((CollationKey) entry.getKey()).getSourceString();
                    List positions = (List) entry.getValue();

                    // start the word index section
                    xw.writeStartElement(XmlConstants.WORD_INDEX_TAG_NAME);

                    // write the word - should be properly escaped
                    xw.writeStartElement(XmlConstants.WORD_TAG_NAME);
                    xw.writeText(word, true);
                    xw.writeEndElement(XmlConstants.WORD_TAG_NAME);

                    // write the word positions
                    for (Iterator j = positions.iterator(); j.hasNext();) {
                        xw.writeStartElement(XmlConstants.POS_TAG_NAME);
                        xw.writeText(((Integer) j.next()).toString());
                        xw.writeEndElement(XmlConstants.POS_TAG_NAME);
                    }

                    // end the word index section
                    xw.writeEndElement(XmlConstants.WORD_INDEX_TAG_NAME);
                }

                // write the contents for last file
                if ((length + xw.getLength()) < limit) {
                    writer.write(xw.getString());
                }

                // write the footer
                writer.write(footer);
                writer.close();
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("Fails to write to document index file", e);
        } finally {
            Util.close(writer);
        }
    }

    /**
     * <p>
     * Generates the header of the document index file. The information of the word source id is listed. This header is
     * included in the beginning of every sequence file.
     * </p>
     *
     * @param documentId the unique id of the document index.
     * @param documentIndex the document index to be persisted.
     *
     * @return the header of the document index file.
     *
     * @throws IndexPersistenceException if any error occurs in extracting information from word source id.
     */
    private String generateDocumentIndexHeader(String documentId, DocumentIndex documentIndex)
        throws IndexPersistenceException {
        WordSourceId wordSourceId = documentIndex.getWordSourceId();
        Locale locale = wordSourceId.getSourceLocale();
        XmlWriter xw = new XmlWriter();

        // start the document index section
        xw.writeStartElement(XmlConstants.DOC_TAG_NAME, XmlConstants.XML_LANG_ATTR_NAME, toXmlLanguage(locale));
        xw.writeStartElement(XmlConstants.DOCUMENT_INDEX_TAG_NAME, XmlConstants.DOCUMENT_ID_ATTR_NAME, documentId);

        // start the word source data section
        xw.writeStartElement(XmlConstants.WORD_SOURCE_DATA_TAG_NAME);

        // write the word source data
        xw.writeStartElement(XmlConstants.SOURCE_IDENTITY_TAG_NAME);
        xw.writeText(Util.compress(wordSourceId.getSourceIdentity()));
        xw.writeEndElement(XmlConstants.SOURCE_IDENTITY_TAG_NAME);
        xw.writeStartElement(XmlConstants.CLASS_NAME_TAG_NAME);
        xw.writeText(wordSourceId.getSourceClassName());
        xw.writeEndElement(XmlConstants.CLASS_NAME_TAG_NAME);
        xw.writeStartElement(XmlConstants.LOCALE_TAG_NAME, toLocaleAttributes(locale));
        xw.writeEndElement(XmlConstants.LOCALE_TAG_NAME);
        xw.writeStartElement(XmlConstants.DELIMITERS_TAG_NAME);

        // write the delimiters - should be properly escaped
        String[] delimiters = wordSourceId.getDelimiters();

        for (int i = 0; i < delimiters.length; i++) {
            xw.writeStartElement(XmlConstants.ENTRY_TAG_NAME, XmlConstants.DELIMITER_ATTR_NAME, delimiters[i], true);
            xw.writeEndElement(XmlConstants.ENTRY_TAG_NAME);
        }

        // end the word source data section
        xw.writeEndElement(XmlConstants.DELIMITERS_TAG_NAME);
        xw.writeEndElement(XmlConstants.WORD_SOURCE_DATA_TAG_NAME);

        return xw.getString();
    }

    /**
     * <p>
     * Generates the footer of the document index file. It includes the closing tags of the xml. This footer is
     * included at the bottom of every sequence file.
     * </p>
     *
     * @return the footer of the document index file.
     */
    private String generateDocumentIndexFooter() {
        XmlWriter xw = new XmlWriter();

        // end the document index section
        xw.writeEndElement(XmlConstants.DOCUMENT_INDEX_TAG_NAME);
        xw.writeEndElement(XmlConstants.DOC_TAG_NAME);

        return xw.getString();
    }

    /**
     * <p>
     * Converts the given locale to the string representation used in the xml:lang attribute.
     * </p>
     *
     * @param locale the locale to convert.
     *
     * @return the value of the xml:lang attribute.
     */
    private String toXmlLanguage(Locale locale) {
        StringBuffer sb = new StringBuffer();
        String country = locale.getCountry();

        sb.append(locale.getLanguage());

        // if country is present, append it as the sublanguage
        if (country.length() > 0) {
            sb.append("-");
            sb.append(country);
        }
        return sb.toString();
    }

    /**
     * <p>
     * Converts the given locale to the attribute map for the locale element in the xml.
     * </p>
     *
     * @param locale the locale to convert.
     *
     * @return the attribute map for the locale element.
     */
    private Map toLocaleAttributes(Locale locale) {
        Map attributes = new HashMap();

        // get the language, country and variant fields
        String language = locale.getLanguage();
        String country = locale.getCountry();
        String variant = locale.getVariant();

        // if any of these is specified, add it as an attribute
        if (language.length() > 0) {
            attributes.put(XmlConstants.LANGUAGE_ATTR_NAME, language);
        }
        if (country.length() > 0) {
            attributes.put(XmlConstants.COUNTRY_ATTR_NAME, country);
        }
        if (variant.length() > 0) {
            attributes.put(XmlConstants.VARIANT_ATTR_NAME, variant);
        }
        return attributes;
    }
}
