/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import com.topcoder.document.index.persistence.impl.Util;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;


/**
 * <p>
 * An xml writer to encapsulate the writing of xml structures, such as element and attribute. It supports the escaping
 * of non-displayable characters, such as the line feed and tab. The output xml is not pretty formatted. The caller is
 * responsible for calling the methods in the right order to ensure a valid xml output.
 * </p>
 *
 * <p>
 * This class and all its methods are package-private, so that they can only be used by classes in the same package.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
class XmlWriter {
    /**
     * <p>
     * The string buffer where the xml output is written.
     * </p>
     */
    private final StringBuffer buffer = new StringBuffer();

    /**
     * <p>
     * Creates a new instance of XmlWriter.
     * </p>
     */
    XmlWriter() {
    }

    /**
     * <p>
     * Writes the start element of the given name with no attributes.
     * </p>
     *
     * @param elementName the name of the element.
     */
    void writeStartElement(String elementName) {
        writeStartElement(elementName, new HashMap());
    }

    /**
     * <p>
     * Writes the start element of the given name with a single attribute.
     * </p>
     *
     * @param elementName the name of the element.
     * @param attrName the name of the attribute.
     * @param attrValue the value of the attribute.
     */
    void writeStartElement(String elementName, String attrName, String attrValue) {
        writeStartElement(elementName, attrName, attrValue, false);
    }

    /**
     * <p>
     * Writes the start element of the given name with a single attribute. The escape flag indicates whether the writer
     * should check for escape characters in the attribute value. If it is true, the attribute value string will be
     * escaped properly.
     * </p>
     *
     * @param elementName the name of the element.
     * @param attrName the name of the attribute.
     * @param attrValue the value of the attribute.
     * @param escape whether the attribute value should be escaped.
     */
    void writeStartElement(String elementName, String attrName, String attrValue, boolean escape) {
        if (escape) {
            attrValue = escape(attrValue);
        }
        writeStartElement(elementName, Collections.singletonMap(attrName, attrValue));
    }

    /**
     * <p>
     * Writes the start element of the given name with multiple attributes.
     * </p>
     *
     * @param elementName the name of the element.
     * @param attributes the mapping from attribute names (String) to attribute values (String).
     */
    void writeStartElement(String elementName, Map attributes) {
        // start the tag
        buffer.append("<");
        buffer.append(elementName);

        // write the attributes inside the tag
        for (Iterator i = attributes.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry = (Map.Entry) i.next();
            String attrName = (String) entry.getKey();
            String attrValue = (String) entry.getValue();

            buffer.append(" ");
            buffer.append(attrName);
            buffer.append("=\"");
            buffer.append(attrValue);
            buffer.append("\"");
        }

        // end the tag
        buffer.append(">");
    }

    /**
     * <p>
     * Writes the end element of the given name.
     * </p>
     *
     * @param elementName the name of the element.
     */
    void writeEndElement(String elementName) {
        buffer.append("<");
        buffer.append("/");
        buffer.append(elementName);
        buffer.append(">");
    }

    /**
     * <p>
     * Writes the specified text contents.
     * </p>
     *
     * @param text the string representing the text contents.
     */
    void writeText(String text) {
        writeText(text, false);
    }

    /**
     * <p>
     * Writes the specified text contents. The escape flag indicates whether the writer should check for escape
     * characters in the text. If it is true, the text string will be escaped properly.
     * </p>
     *
     * @param text the string representing the text contents.
     * @param escape whether the text should be escaped.
     */
    void writeText(String text, boolean escape) {
        if (escape) {
            text = escape(text);
        }
        buffer.append(text);
    }

    /**
     * <p>
     * Returns the string written so far.
     * </p>
     *
     * @return the string written so far.
     */
    String getString() {
        return buffer.toString();
    }

    /**
     * <p>
     * Returns the length of the string written so far, in bytes.
     * </p>
     *
     * @return the length of the string written so far.
     */
    int getLength() {
        // no need to take care of double-byte characters since they must be escaped
        return buffer.length();
    }

    /**
     * <p>
     * Resets the writer such that it has no output string.
     * </p>
     */
    void reset() {
        buffer.setLength(0);
    }

    /**
     * <p>
     * Escapes the given string properly. If any character is neither a letter nor a digit, it will be escaped using
     * the xml representation method.
     * </p>
     *
     * @param str the string to be escaped.
     *
     * @return the escaped string.
     */
    private String escape(String str) {
        StringBuffer sb = new StringBuffer();

        for (int i = 0; i < str.length(); i++) {
            char ch = str.charAt(i);

            if (Character.isLetterOrDigit(ch)) {
                // if the character is a letter or a digit, no need to escape
                sb.append(ch);
            } else {
                // otherwise use the hex string representation
                sb.append("&#x");
                sb.append(Util.toHexString(ch));
                sb.append(";");
            }
        }
        return sb.toString();
    }
}
