/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.wordsource.WordSourceId;

import junit.framework.TestCase;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;


/**
 * The base test class for the IndexPersistence implementations. The tests are based on the contract defined by the
 * interface. Basically it performs various argument checks.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public abstract class BaseIndexPersistenceTest extends TestCase {
    /**
     * The IndexPersistence instance to test against.
     */
    private IndexPersistence persistence = null;

    /**
     * The word source id argument supplied to the methods.
     */
    protected WordSourceId wordSourceId = null;

    /**
     * The document index argument supplied to the methods.
     */
    protected DocumentIndex documentIndex = null;

    /**
     * The collection id argument supplied to the methods.
     */
    protected String collectionId = null;

    /**
     * The collection index argument supplied to the methods.
     */
    protected CollectionIndex collectionIndex = null;

    /**
     * The unique id of the document whose index is to be persisted.
     */
    protected String documentId = null;

    /**
     * The unique ids of the documents whose indexes are to be persisted.
     */
    protected String[] documentIds = null;

    /**
     * Prepares the method arguments for testing.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void setUp() throws Exception {
        // prepare the arguments for document index
        wordSourceId = TestHelper.createWordSourceId();
        documentIndex = TestHelper.createDocumentIndex();

        // prepare the arguments for collection index
        collectionId = "collection";
        collectionIndex = TestHelper.createCollectionIndex();

        documentId = "d87cbcf233a87e8b4a7f3b76f0009d1c";
        documentIds = new String[] {
                "78d40f9ab62b2e85f658b42e0abadd64", "eab5e77b4394b21801069ed1d0b327bf",
                "b0a04a018e9bac0666dac6bbcddc819c"
            };
    }

    /**
     * Test of addDocumentIndex method with null document index. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddDocumentIndex_NullDocumentIndex()
        throws Exception {
        try {
            persistence.addDocumentIndex(null);
            fail("Adds null document index");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of addDocumentIndex method with bad document index (non-serializable source identity). Expects
     * IndexPersistenceException.
     */
    public void testAddDocumentIndex_BadDocumentIndex() {
        try {
            persistence.addDocumentIndex(TestHelper.createIllegalDocumentIndex());
            fail("Adds a bad document index");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of addDocumentIndex method with existing document index. Expects IndexPersistenceException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddDocumentIndex_ExistDocumentIndex()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        try {
            persistence.addDocumentIndex(documentIndex);
            fail("Adds an existing document index");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of getDocumentIndex method with null word source id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetDocumentIndex_NullWordSourceId()
        throws Exception {
        try {
            persistence.getDocumentIndex(null);
            fail("Gets document index with null word source id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of getDocumentIndex method with bad word source id (non-serializable source identity). Expects
     * IndexPersistenceException.
     */
    public void testGetDocumentIndex_BadWordSourceId() {
        try {
            persistence.getDocumentIndex(TestHelper.createIllegalWordSourceId());
            fail("Gets document index with bad word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of getDocumentIndex method with non-existing word source id. Should return null.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetDocumentIndex_NonExistWordSourceId()
        throws Exception {
        assertNull("Returns an incorrect document index", persistence.getDocumentIndex(wordSourceId));
    }

    /**
     * Test of getDocumentIndex method with existing word source id. Verifies if the fields of the returned document
     * index were populated.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetDocumentIndex_ExistWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        DocumentIndex documentIndex = persistence.getDocumentIndex(wordSourceId);
        Map words = TestHelper.createDocumentIndexWords();

        assertEquals("Returns an incorrect document index", wordSourceId, documentIndex.getWordSourceId());
        assertEquals("Returns an incorrect document index", words, documentIndex.getWords());
    }

    /**
     * Test of removeDocumentIndex method with null word source id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveDocumentIndex_NullWordSourceId()
        throws Exception {
        try {
            persistence.removeDocumentIndex(null);
            fail("Removes document index with null word source id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of removeDocumentIndex method with bad word source id (non-serializable source identity). Expects
     * IndexPersistenceException.
     */
    public void testRemoveDocumentIndex_BadWordSourceId() {
        try {
            persistence.removeDocumentIndex(TestHelper.createIllegalWordSourceId());
            fail("Removes document index with bad word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of removeDocumentIndex method with non-existing word source id. Expects IndexPersistenceException.
     */
    public void testRemoveDocumentIndex_NonExistWordSourceId() {
        try {
            persistence.removeDocumentIndex(wordSourceId);
            fail("Removes document index with non-existing word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of removeDocumentIndex method with used word source id (use count is non-zero). Expects
     * IndexPersistenceException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveDocumentIndex_UsedWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        // make sure it is in use
        persistence.increaseDocumentUseCount(wordSourceId);

        try {
            persistence.removeDocumentIndex(wordSourceId);
            fail("Removes document index with used word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of addCollectionIndex method with null collection index. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddCollectionIndex_NullCollectionIndex()
        throws Exception {
        try {
            persistence.addCollectionIndex(null);
            fail("Adds null collection index");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of addCollectionIndex method with bad collection index (non-persisted document index). Expects
     * IndexPersistenceException.
     */
    public void testAddCollectionIndex_BadCollectionIndex() {
        try {
            persistence.addCollectionIndex(collectionIndex);
            fail("Adds bad collection index");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of addCollectionIndex method with existing collection index. Expects IndexPersistenceException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddCollectionIndex_ExistCollectionIndex()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        try {
            persistence.addCollectionIndex(collectionIndex);
            fail("Adds an existing collection index");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of getCollectionIndex method with null collection id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetCollectionIndex_NullCollectionId()
        throws Exception {
        try {
            persistence.getCollectionIndex(null);
            fail("Gets collection index with null collection id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of getCollectionIndex method with empty collection id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetCollectionIndex_EmptyCollectionId()
        throws Exception {
        try {
            persistence.getCollectionIndex("");
            fail("Gets collection index with empty collection id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of getCollectionIndex method with non-existing collection id. Should return null.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetCollectionIndex_NonExistCollectionId()
        throws Exception {
        assertNull("Returns an incorrect collection index", persistence.getCollectionIndex(collectionId));
    }

    /**
     * Test of getCollectionIndex method with existing collection id. Verifies if the fields of the returned collection
     * index were populated.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetCollectionIndex_ExistCollectionId()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        CollectionIndex collectionIndex = persistence.getCollectionIndex(collectionId);
        Map allWords = TestHelper.createCollectionIndexWords();
        Set allDocumentIds = TestHelper.createCollectionIndexDocumentIds();

        assertNull("Returns an incorrect collection index", collectionIndex.getOwner());
        assertEquals("Returns an incorrect collection index", allWords, collectionIndex.getWords());
        assertEquals("Returns an incorrect collection index", allDocumentIds, collectionIndex.getAllDocumentIds());
        assertEquals("Returns an incorrect collection index", collectionId, collectionIndex.getId());
    }

    /**
     * Test of removeCollectionIndex method with null collection id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveCollectionIndex_NullCollectionId()
        throws Exception {
        try {
            persistence.removeCollectionIndex(null);
            fail("Removes collection index with null collection id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of removeCollectionIndex method with empty collection id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveCollectionIndex_EmptyCollectionId()
        throws Exception {
        try {
            persistence.removeCollectionIndex("");
            fail("Removes collection index with empty collection id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of removeCollectionIndex method with non-existing collection id. Expects IndexPersistenceException.
     */
    public void testRemoveCollectionIndex_NonExistCollectionId() {
        try {
            persistence.removeCollectionIndex(collectionId);
            fail("Removes collection index with non-existing collection id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of updateCollectionIndex method with null collection index. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testUpdateCollectionIndex_NullCollectionIndex()
        throws Exception {
        try {
            persistence.updateCollectionIndex(null);
            fail("Updates null collection index");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of updateCollectionIndex method with non-existing collection index. Expects IndexPersistenceException.
     */
    public void testUpdateCollectionIndex_NonExistCollectionIndex() {
        try {
            persistence.updateCollectionIndex(collectionIndex);
            fail("Updates a non-existing collection index");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of updateCollectionIndex method with bad collection index (non-persisted document index). Expects
     * IndexPersistenceException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testUpdateCollectionIndex_BadCollectionIndex()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        // make sure the new document index is non-persisted
        wordSourceIds.clear();
        wordSourceIds.add(wordSourceId);
        collectionIndex = new CollectionIndex(null, new HashMap(), wordSourceIds, collectionId);

        try {
            persistence.updateCollectionIndex(collectionIndex);
            fail("Updates bad collection index");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of increaseDocumentUseCount method with null word source id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testIncreaseDocumentUseCount_NullWordSourceId()
        throws Exception {
        try {
            persistence.increaseDocumentUseCount(null);
            fail("Increases document use count with null word source id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of increaseDocumentUseCount method with non-existing word source id. Expects IndexPersistenceException.
     */
    public void testIncreaseDocumentUseCount_NonExistWordSourceId() {
        try {
            persistence.increaseDocumentUseCount(wordSourceId);
            fail("Increases document use count with non-existing word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of decreaseDocumentUseCount method with null word source id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testDecreaseDocumentUseCount_NullWordSourceId()
        throws Exception {
        try {
            persistence.decreaseDocumentUseCount(null);
            fail("Decreases document use count with null word source id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of decreaseDocumentUseCount method with non-existing word source id. Expects IndexPersistenceException.
     */
    public void testDecreaseDocumentUseCount_NonExistWordSourceId() {
        try {
            persistence.decreaseDocumentUseCount(wordSourceId);
            fail("Decreases document use count with non-existing word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of getIndexedDocuments method with empty document index. Should return an empty set.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetIndexedDocuments_Empty() throws Exception {
        assertTrue("Returns an incorrect set of document IDs", persistence.getIndexedDocuments().isEmpty());
    }

    /**
     * Test of getIndexedDocuments method with non-empty document index. Should return the set of all word source ids
     * in the persistence.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetIndexedDocuments_NonEmpty() throws Exception {
        // make sure non-empty document index
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        assertEquals("Returns an incorrect set of document IDs", wordSourceIds, persistence.getIndexedDocuments());
    }

    /**
     * Test of getDocumentUseCount method with null word source id. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetDocumentUseCount_NullWordSourceId()
        throws Exception {
        try {
            persistence.getDocumentUseCount(null);
            fail("Gets document use count with null word source id");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of getDocumentUseCount method with non-existing word source id. Expects IndexPersistenceException.
     */
    public void testGetDocumentUseCount_NonExistWordSourceId() {
        try {
            persistence.getDocumentUseCount(wordSourceId);
            fail("Gets document use count with non-existing word source id");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Test of getDocumentUseCount method with existing word source id. Verifies if the correct use count is returned.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetDocumentUseCount_ExistWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        assertEquals("Returns an incorrect use count", 0, persistence.getDocumentUseCount(wordSourceId));
    }

    /**
     * Sets the IndexPersistence instance to test against. This should be called in the setUp() method of subclass.
     *
     * @param persistence the IndexPersistence instance to test against.
     */
    protected void setPersistence(IndexPersistence persistence) {
        this.persistence = persistence;
    }
}
