/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.IndexManager;
import com.topcoder.document.index.persistence.impl.db.FastAccessDBIndexPersistence;
import com.topcoder.document.index.persistence.impl.db.SimpleAccessDBIndexPersistence;
import com.topcoder.document.index.persistence.impl.xml.XmlHelper;
import com.topcoder.document.index.persistence.impl.xml.XmlIndexPersistence;
import com.topcoder.document.index.wordsource.WordSourceId;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;


/**
 * A simple demonstration of the Document Indexer Persistence component.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class DemoTest extends TestCase {
    /**
     * Creates a test suite for the tests in this test case.
     *
     * @return a TestSuite for this test case
     */
    public static Test suite() {
        TestSuite suite = new TestSuite(DemoTest.class);

        return suite;
    }

    /**
     * Sets up the namespaces and file resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void setUp() throws Exception {
        ConfigHelper.loadConfig();
        XmlHelper.setupDir();
    }

    /**
     * Clears all the namespaces and file resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void tearDown() throws Exception {
        XmlHelper.cleanupDir();
        ConfigHelper.unloadConfig();
    }

    /**
     * Demonstrates how to plug the IndexPersistence instance to the IndexManager. In this way, all the document
     * indexes and collection indexes managed by the IndexManager can get updated in the persistence through the
     * method calls of Document Indexer. This is probably the most common usage from the user point of view.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testUsage() throws Exception {
        // Create an xml based persistence
        IndexPersistence xmlPersistence = new XmlIndexPersistence(ConfigHelper.XML_NAMESPACE);

        // pass it to the manager
        IndexManager manager = new IndexManager(xmlPersistence);

        // Create a fast access db based persistence
        IndexPersistence fastDBPersistence = new FastAccessDBIndexPersistence(ConfigHelper.DB_NAMESPACE);

        // pass it to the manager
        manager = new IndexManager(fastDBPersistence);

        // Create a simple access db based persistence
        IndexPersistence simpleDBPersistence = new SimpleAccessDBIndexPersistence(ConfigHelper.DB_NAMESPACE);

        // pass it to the manager
        manager = new IndexManager(simpleDBPersistence);
    }

    /**
     * Shows the CRUD operations for document index and collection index. Here we use xml persistence as an example,
     * the database persistence is similar. Normally the APIs would not be called directly by the user. They are used
     * by the classes defined in the Document Indexer component.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testPersistence() throws Exception {
        // create some objects used by the methods
        WordSourceId wordSourceId = TestHelper.createWordSourceId();
        DocumentIndex documentIndex = TestHelper.createDocumentIndex();
        CollectionIndex collectionIndex = TestHelper.createCollectionIndex();

        // Create an xml based persistence
        IndexPersistence xmlPersistence = new XmlIndexPersistence(ConfigHelper.XML_NAMESPACE);

        // add a document index to the persistence
        xmlPersistence.addDocumentIndex(documentIndex);

        // get the document index back
        documentIndex = xmlPersistence.getDocumentIndex(wordSourceId);

        // make sure the doucment indexes of collection index are persisted
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            xmlPersistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // add a collection index to the persistence
        xmlPersistence.addCollectionIndex(collectionIndex);

        // get the collection index back
        collectionIndex = xmlPersistence.getCollectionIndex("collection");

        // update the collection index
        xmlPersistence.updateCollectionIndex(collectionIndex);

        // remove the collection index
        xmlPersistence.removeCollectionIndex("collection");

        // increase the usage count of document index
        xmlPersistence.increaseDocumentUseCount(wordSourceId);

        // decrease the usage count of document index
        xmlPersistence.decreaseDocumentUseCount(wordSourceId);

        // get the usage count of document index
        int usageCount = xmlPersistence.getDocumentUseCount(wordSourceId);

        // get all the persisted document indexes
        Set documentIds = xmlPersistence.getIndexedDocuments();

        // remove the document index
        xmlPersistence.removeDocumentIndex(wordSourceId);
    }
}
