/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.BaseIndexPersistenceTest;
import com.topcoder.document.index.persistence.ConfigHelper;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * The base test class for the AbstractDBIndexPersistence implementations. This class assumes that the DOCUMENT,
 * DELIMITER, DOCUMENT_COLLECTION and DOCUMENT_COLLECTION_DOCUMENT_XREF tables are common to the database persistence.
 * If not, the test cases should be detached from this class.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public abstract class BaseDBIndexPersistenceTest extends BaseIndexPersistenceTest {
    /**
     * The sql statement to check a record in the DOCUMENT table.
     */
    private static final String SQL_SEL_DOCUMENT = "SELECT DOCUMENT_ID FROM DOCUMENT WHERE DOCUMENT_ID = ? "
        + "AND LOCALE = ? AND CLASS_NAME = ? AND USE_COUNT = ?";

    /**
     * The sql statement to check a record in the DELIMITER table.
     */
    private static final String SQL_SEL_DELIMITER = "SELECT DOCUMENT_ID FROM DELIMITER WHERE DOCUMENT_ID = ? "
        + "AND DELIMITER_SORT_ORDER = ?";

    /**
     * The sql statement to check a record in the DOCUMENT_COLLECTION table.
     */
    private static final String SQL_SEL_COLLECTION = "SELECT COLLECTION_ID FROM DOCUMENT_COLLECTION "
        + "WHERE COLLECTION_ID = ?";

    /**
     * The sql statement to check a record in the DOCUMENT_COLLECTION_DOCUMENT_XREF table.
     */
    private static final String SQL_SEL_DOCUMENT_COLLECTION = "SELECT COLLECTION_ID "
        + "FROM DOCUMENT_COLLECTION_DOCUMENT_XREF WHERE COLLECTION_ID = ? AND DOCUMENT_ID = ?";

    /**
     * The parameters to set on the sql query.
     */
    protected List parameters = null;

    /**
     * The AbstractDBIndexPersistence instance to test against.
     */
    private AbstractDBIndexPersistence persistence = null;

    /**
     * Sets up the namespaces and database resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void setUp() throws Exception {
        super.setUp();
        ConfigHelper.loadConfig();
        DBHelper.clearTables();

        // prepare the parameters for checking the existence of a record
        parameters = new ArrayList();
    }

    /**
     * Clears all the namespaces and database resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void tearDown() throws Exception {
        DBHelper.clearTables();
        ConfigHelper.unloadConfig();
        super.tearDown();
    }

    /**
     * Test of addDocumentIndex method with non-existing document index. Verifies if the document index was persisted
     * after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddDocumentIndex_NonExistDocumentIndex()
        throws Exception {
        persistence.addDocumentIndex(documentIndex);

        // check the DOCUMENT table
        parameters.add(documentId);
        parameters.add("en");
        parameters.add("className");
        parameters.add(new Integer(0));
        assertTrue("Fails to add document index to DOCUMENT table",
            DBHelper.exist(SQL_SEL_DOCUMENT, parameters.toArray()));

        // check the DELIMITER table
        String[] delimiters = {"\t", "\n", "\r"};

        for (int i = 0; i < delimiters.length; i++) {
            parameters.clear();
            parameters.add(documentId);
            parameters.add(new Integer(i));
            assertTrue("Fails to add document index to DELIMITER table",
                DBHelper.exist(SQL_SEL_DELIMITER, parameters.toArray()));
        }
    }

    /**
     * Test of removeDocumentIndex method with unused word source id (use count is zero). Verifies if the document
     * index was removed from persistence after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveDocumentIndex_UnusedWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        persistence.removeDocumentIndex(wordSourceId);

        // check the DOCUMENT table
        assertTrue("Fails to remove document index from DOCUMENT table", DBHelper.isEmptyTable("DOCUMENT"));

        // check the DELIMITER table
        assertTrue("Fails to remove document index from DELIMITER table", DBHelper.isEmptyTable("DELIMITER"));
    }

    /**
     * Test of addCollectionIndex method with non-existing collection index. Verifies if the collection index was
     * persisted after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddCollectionIndex_NonExistCollectionIndex()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        persistence.addCollectionIndex(collectionIndex);

        // check the COLLECTION table
        parameters.add(collectionId);
        assertTrue("Fails to add collection index to COLLECTION table",
            DBHelper.exist(SQL_SEL_COLLECTION, parameters.toArray()));

        // check the DOCUMENT_COLLECTION_DOCUMENT_XREF table      
        for (int i = 0; i < documentIds.length; i++) {
            parameters.clear();
            parameters.add(collectionId);
            parameters.add(documentIds[i]);
            assertTrue("Fails to add collection index to DOCUMENT_COLLECTION_DOCUMENT_XREF table",
                DBHelper.exist(SQL_SEL_DOCUMENT_COLLECTION, parameters.toArray()));
        }
    }

    /**
     * Test of removeCollectionIndex method with existing collection id. Verifies if the collection index was removed
     * from persistence after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveCollectionIndex_ExistCollectionId()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        persistence.removeCollectionIndex(collectionId);

        // check the DOCUMENT_COLLECTION table
        assertTrue("Fails to remove document index from DOCUMENT_COLLECTION table",
            DBHelper.isEmptyTable("DOCUMENT_COLLECTION"));

        // check the DOCUMENT_COLLECTION_DOCUMENT_XREF table
        assertTrue("Fails to remove document index from DOCUMENT_COLLECTION_DOCUMENT_XREF table",
            DBHelper.isEmptyTable("DOCUMENT_COLLECTION_DOCUMENT_XREF"));
    }

    /**
     * Test of updateCollectionIndex method with existing collection index. Verifies if the collection index was
     * updated in the persistence after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testUpdateCollectionIndex_ExistCollectionIndex()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        // make sure the new document index is persisted
        persistence.addDocumentIndex(documentIndex);

        wordSourceIds.clear();
        wordSourceIds.add(wordSourceId);
        collectionIndex = new CollectionIndex(null, new HashMap(), wordSourceIds, collectionId);
        persistence.updateCollectionIndex(collectionIndex);

        // verify if the old records are removed      
        for (int i = 0; i < documentIds.length; i++) {
            parameters.clear();
            parameters.add(collectionId);
            parameters.add(documentIds[i]);
            assertFalse("Fails to update collection index in DOCUMENT_COLLECTION_DOCUMENT_XREF table",
                DBHelper.exist(SQL_SEL_DOCUMENT_COLLECTION, parameters.toArray()));
        }

        // verify if the new records are added
        parameters.clear();
        parameters.add(collectionId);
        parameters.add(documentId);
        assertTrue("Fails to update collection index in DOCUMENT_COLLECTION_DOCUMENT_XREF table",
            DBHelper.exist(SQL_SEL_DOCUMENT_COLLECTION, parameters.toArray()));
    }

    /**
     * Test of increaseDocumentUseCount method with existing word source id. Verifies if the use count was updated in
     * the persistence after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testIncreaseDocumentUseCount_ExistWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        persistence.increaseDocumentUseCount(wordSourceId);

        parameters.add(documentId);
        parameters.add("en");
        parameters.add("className");
        parameters.add(new Integer(1));
        assertTrue("Fails to increase use count in DOCUMENT table",
            DBHelper.exist(SQL_SEL_DOCUMENT, parameters.toArray()));
    }

    /**
     * Test of decreaseDocumentUseCount method with existing word source id. Verifies if the use count was updated in
     * the persistence after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testDecreaseDocumentUseCount_ExistWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        persistence.decreaseDocumentUseCount(wordSourceId);

        parameters.add(documentId);
        parameters.add("en");
        parameters.add("className");
        parameters.add(new Integer(-1));
        assertTrue("Fails to decrease use count in DOCUMENT table",
            DBHelper.exist(SQL_SEL_DOCUMENT, parameters.toArray()));
    }

    /**
     * Sets the AbstractDBIndexPersistence instance to test against. This should be called in the setUp() method of
     * subclass.
     *
     * @param persistence the AbstractDBIndexPersistence instance to test against.
     */
    protected void setPersistence(AbstractDBIndexPersistence persistence) {
        super.setPersistence(persistence);
        this.persistence = persistence;
    }
}
