/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;

import com.topcoder.document.index.persistence.ConfigHelper;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;


/**
 * A helper class to perform those database operations for testing. These include checking whether a table is empty,
 * clearing the tables, etc.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class DBHelper {
    /**
     * Private constructor to prevent instantiation.
     */
    private DBHelper() {
    }

    /**
     * Gets the connection to perform database operations. This method assumes that the namespace of the db connection
     * factory has been loaded.
     *
     * @return the database connection
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public static Connection getConnection() throws Exception {
        DBConnectionFactory factory = new DBConnectionFactoryImpl(ConfigHelper.CONNECTION_NAMESPACE);

        return factory.createConnection(ConfigHelper.CONNECTION_PRODUCER_NAME);
    }

    /**
     * Checks whether the table of the given name is empty. An empty table means it has no rows.
     *
     * @param tableName the name of the table
     *
     * @return whether the table of the given name is empty
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public static boolean isEmptyTable(String tableName)
        throws Exception {
        String sql = "SELECT * FROM " + tableName;

        Connection conn = getConnection();
        PreparedStatement pstmt = conn.prepareStatement(sql);
        ResultSet rs = pstmt.executeQuery();

        boolean result = !rs.next();

        rs.close();
        pstmt.close();
        conn.close();

        return result;
    }

    /**
     * Checks whether the record exists in the table. This is determined by executing the sql query with the specified
     * parameters. The sql query should be a SELECT statement.
     *
     * @param sql the sql query to execute.
     * @param parameters an array of parameters.
     *
     * @return whether the record exists in the table.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public static boolean exist(String sql, Object[] parameters)
        throws Exception {
        Connection conn = getConnection();
        PreparedStatement pstmt = conn.prepareStatement(sql);

        for (int i = 0; i < parameters.length; i++) {
            pstmt.setObject(i + 1, parameters[i]);
        }

        ResultSet rs = pstmt.executeQuery();
        boolean result = rs.next();

        rs.close();
        pstmt.close();
        conn.close();

        return result;
    }

    /**
     * Clears all the tables in the database.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public static void clearTables() throws Exception {
        String[] tableNames = {
                "DOCUMENT_COLLECTION_DOCUMENT_XREF", "DELIMITER", "WORD_POSITION", "WORD", "WORD_DATA",
                "DOCUMENT_COLLECTION", "DOCUMENT"
            };

        Connection conn = getConnection();

        for (int i = 0; i < tableNames.length; i++) {
            String sql = "DELETE FROM " + tableNames[i];
            PreparedStatement pstmt = conn.prepareStatement(sql);

            pstmt.executeUpdate();
            pstmt.close();
        }

        conn.close();
    }
}
